unit NEXLowLevel;

interface

uses
  {$IFDEF LINUX}
  LibC,
  {$ENDIF}
  {$IFDEF MSWINDOWS}
  SmallPort, Windows,
  {$ENDIF}
  NEXTypes, Sysutils;

function InitPlayer1 : uint16;
function InitPlayer2 : uint16;
function SelectBank(BankNum : uint8) : uint16;
function ReadFlashRam(BankNum : uint8; SectorNum : uint16; NumBytes : sint32;
  DataPtr : pointer) : boolean;
function PrepareWriteFlashRam(BankNum : uint8; ChunkNum : uint16;
  AParam : sint32) : uint8;
function WriteFlashRam(BankNum : uint8; SectorNum : uint16; NumBytes : sint32;
  DataPtr : pointer) : integer;
procedure KickPlayer;

{
Low-level I/O procedures (<..> indicates not written yet):
  401660    <InitPlayerMemory1>
  406C50    <InitPlayerMemory2>
  409F30    InitPlayer1
  409FE0    PlayerCmd1
  40A0D0    ReadFlashRam with count=$210
  40A230    ReadFlashRam
  40A390    Preparing flash for writing
  40A570    Writing $210 bytes
  40A750    Writing an arbitrary number of bytes

Possible memory layout: 4 banks of 1024 chunks of 32 sectors of 528 bytes
Each FAT block is 2 chunks (~32kb)
}

implementation

{$IFDEF MSWINDOWS}
var
  LastKick    : cardinal = 0;
  LastKick64  : int64 = 0;

procedure KickPlayer;
var
  ThisKick64  : int64;
  CounterFreq : int64;
begin
  ThisKick64 := 0;
  if not QueryPerformanceCounter(ThisKick64) then
  begin
    // No high-resolution counter, fall back to gettickcount
    if GetTickCount - LastKick > 64 then
    begin
      InitPlayer2;
      LastKick := GetTickCount;
    end;
  end
  else
  begin
    QueryPerformanceFrequency(CounterFreq);
    if (ThisKick64 - LastKick64)/CounterFreq > (64/1000) then
    begin
      InitPlayer2;
      LastKick64 := ThisKick64;
    end;
  end;
end;
{$ENDIF}

{$IFDEF LINUX}
var
  LastKick    : TTimeVal = (tv_sec:0;tv_usec:0);

procedure KickPlayer;
var
  ThisKick    : TTimeVal;
  ElapsedTime : TTimeVal;
begin
  GetTimeOfDay(ThisKick, nil);
  ElapsedTime := timersub(ThisKick, LastKick);
  if (ElapsedTime.tv_sec > 0) or (ElapsedTime.tv_usec > 64000) then
  begin
    InitPlayer2;
    LastKick := ThisKick;
  end;
end;

// PortNum = eax, PortData = edx
procedure OutPortB(PortNum : uint16; PortData : uint8);
asm
  xchg  eax,edx
  out   dx,al
  xchg  eax,edx
end;

// PortNum = eax
function InPortB(PortNum : uint16) : uint8;
asm
  push  edx
  mov   edx,eax
  xor   eax,eax
  in    al,dx
  pop   edx
end;
{$ENDIF}

{$IFDEF MSWINDOWS}
var
  PortObj : TSmallPort;

procedure OutPortB(PortNum : uint16; PortData : uint8);
begin
  PortObj.WriteByte(PortNum, PortData);
end;

function InPortB(PortNum : uint16) : uint8;
begin
  Result := PortObj.ReadByte(PortNum);
end;

{$ENDIF}
procedure SendByte08(DataByte : uint8);
begin
  OutPortB($378, DataByte);
  OutPortB($37A, $08);
  OutPortB($37A, $00);
end;

procedure SendByte02(DataByte : uint8);
begin
  OutPortB($378, DataByte);
  OutPortB($37A, $02);
  OutPortB($37A, $00);
end;

function RecvByte02 : uint8;
var
  TmpDat  : uint8;
begin
  OutPortB($37A, $02);
  TmpDat := InPortB($379);
  OutPortB($37A, $00);
  Result := (InPortB($379) and $F0) or (TmpDat shr 4);
end;

// 00409F30
function InitPlayer1 : uint16;
var
  TmpByte : uint8;
begin
  SendByte08($C0);
  TmpByte := RecvByte02;
  SendByte08($D0);
  SendByte02($85);
  SendByte08($D0);
  SendByte02($80);
  SendByte08($C8);
  Result := (RecvByte02 shl 8) or (TmpByte);
end;

// $00409EC0
// Return values:
//  $02xx = Init failed (xx = code from player)
//  $00xx = Init success (xx = code from player)
// Player codes:
//  $55   = OK/Ready?
function InitPlayer2 : uint16;
var
  NumTries    : integer;
  LastResult  : uint16;
begin
  for NumTries := 0 to 10 do
  begin
    LastResult := InitPlayer1;
    if (LastResult and $0F00) = $0F00 then
    begin
      Result := LastResult and $00FF;
      exit;
    end;
    sleep(100);
  end;

  Result := (LastResult and $00FF) or $0200;
end;

// 00409FE0
// Seems to return $ECxx for valid banks.
//   xx = $E3, $E5, $E6 means 16 sectors per chunk
//   xx = not the above means 32 sectors per chunk
function SelectBank(BankNum : uint8) : uint16;
begin
  SendByte08($D0);
  SendByte02((BankNum and $07) or $40);

  SendByte08($DD);
  SendByte02($90);

  SendByte08($DE);
  SendByte02($00);

  SendByte08($D8);
  Result := RecvByte02 shl 8;
  Result := Result or RecvByte02;

  SendByte08($D0);
  SendByte02($00);
end;

// 0040A230
// 0040A0D0 the same but NumBytes fixed at $210
function ReadFlashRam(BankNum : uint8; SectorNum : uint16; NumBytes : sint32;
  DataPtr : pointer) : boolean;
var
  StartTime : TDateTime;
  PlayerRdy : boolean;
  ByteLoop  : integer;
begin
  SendByte08($D0);
  SendByte02((BankNum and $07) or $40);

  SendByte08($DD);
  SendByte02($00);

  SendByte08($DE);
  SendByte02($00);
  SendByte02(SectorNum);
  SendByte02(SectorNum shr 8);

  SendByte08($D8);

  // Give 3 seconds to signal "ready"
  StartTime := GetTime;
  PlayerRdy := false;
  while (GetTime - StartTime) < 3/(24*60*60) do
  begin
    if (InPortB($379) and $08) = $08 then
    begin
      PlayerRdy := true;
      break;
    end;
  end;
  if (NumBytes > 0) and PlayerRdy then
  begin
    for ByteLoop := 0 to NumBytes - 1 do
    begin
      PByteArray(DataPtr)^[ByteLoop] := RecvByte02;
    end;
  end;

  SendByte08($D0);
  SendByte02($00);

  Result := PlayerRdy;

  NexDoLog('ReadFlashRam(' + IntToHex(BankNum, 2) + ', ' +
    IntToStr(SectorNum) + ', ' + IntToStr(NumBytes) + ', <ptr>)');
end;

// 0040A390
// AParam = $73 on my player
function PrepareWriteFlashRam(BankNum : uint8; ChunkNum : uint16;
  AParam : sint32) : uint8;
var
  PlayerRdy : boolean;
  TmpByte   : byte;
  ALoop     : integer;
  StartTime : TDateTime;
begin
  for ALoop := 0 to 10 do
  begin
    SendByte08($D0);
    SendByte02((BankNum and $07) or $40);
    SendByte08($DD);
    SendByte02($60);
    SendByte08($DE);

    if (AParam = $E3) or (AParam = $E5) or (AParam = $E6) then
    begin
      // 16 sectors per chunk (~8kb)
      SendByte02(ChunkNum shl 4);
      SendByte02(ChunkNum shr 4);
    end
    else
    begin
      // 32 sectors per chunk (~16kb)
      SendByte02(ChunkNum shl 5);
      SendByte02(ChunkNum shr 3);
    end;


    SendByte08($DD);
    SendByte02($D0);

    PlayerRdy := false;
    StartTime := GetTime;

    while ((GetTime - StartTime) < 3/(24*60*60)) and (not PlayerRdy) do
      if (InPortB($379) and $08) = $08 then PlayerRdy := true;

    SendByte02($70);
    SendByte08($D8);

    TmpByte := RecvByte02;

    SendByte08($D0);
    SendByte02($00);
    if PlayerRdy then break;
  end;

  if PlayerRdy then
    Result := TmpByte and $FD
  else
    Result := TmpByte or $02;
end;

// 0040A570
// 0040A750 the same but fixed NumBytes = $210
function WriteFlashRam(BankNum : uint8; SectorNum : uint16; NumBytes : sint32;
  DataPtr : pointer) : integer;
var
  ALoop     : integer;
  PlayerRdy : boolean;
  SendLoop  : integer;
  StartTime : TDateTime;
  AByte     : integer;
begin
  for ALoop := 0 to 10 do
  begin
    SendByte08($D0);
    SendByte02((BankNum and $07) or $40);
    SendByte08($DD);
    SendByte02($80);
    SendByte08($DE);
    SendByte02($00);
    SendByte02(SectorNum);
    SendByte02(SectorNum shr 8);
    SendByte08($DC);

    for SendLoop := 0 to $210 - 1 do
      SendByte02(PByteArray(DataPtr)^[SendLoop]);

    SendByte08($DD);
    SendByte02($10);

    PlayerRdy := false;
    StartTime := GetTime;
    while ((GetTime - StartTime) < 3/(24*60*60)) and (not PlayerRdy) do
      if (InPortB($379) and $08) = $08 then PlayerRdy := true;

    SendByte02($70);
    SendByte08($D8);
    AByte := RecvByte02;
    SendByte08($D0);
    SendByte02($00);
    if PlayerRdy then break;
  end;

  if not PlayerRdy then
    Result := AByte and $FD
  else
    Result := AByte or $02;

  NexDoLog('WriteFlashRam(' + IntToHex(BankNum, 2) + ', ' +
    IntToStr(SectorNum) + ', ' + IntToStr(NumBytes) + ', <ptr>)');
end;

initialization
{$IFDEF LINUX}
  ioperm($378, 3, 1);
{$ENDIF}
{$IFDEF MSWINDOWS}
  PortObj := TSmallPort.Create;
  PortObj.Opened := true;
{$ENDIF}

finalization
{$IFDEF MSWINDOWS}
  PortObj.Free;
{$ENDIF}

end.
