unit NEXObjs;

interface

uses
  SysUtils, Classes, NexUtils, NexTypes, NexLowLevel;

type
  TNexFilesystem = class(TObject)
  private
    FNexMemory  : TNEXLowMemory;
    FOrigMem    : TNEXLowMemory;
  protected
  public
    constructor Create;
    destructor Destroy; override;

    function AllocFileBlock(AfterBlock : integer) : integer;
    procedure FreeFileBlock(BlockNum : integer);
    function NextBlock(CurBlock : integer) : integer;
    function PrevBlock(CurBlock : integer) : integer;

    function AllocFileEntry : PFileEntry;
    procedure FreeFileEntry(FE : PFileEntry);
    function GetFileEntry(FileNum : integer) : PFileEntry;

    procedure Commit;
    function BytesToCommit : integer;
    procedure Revert;

    function FreeBytes : integer;
    function TotalBytes : integer;
  end;

  TNexStream = class(TStream)
  private
    FFileSys    : TNexFilesystem;
    FFileEntry  : PFileEntry;
    FCurPos     : integer;
  protected
    procedure SetSize(NewSize: Longint); override;
  public
    constructor Create(nfs : TNexFilesystem; SongNum : integer);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    property FileEntry : PFileEntry read FFileEntry;
  end;

implementation

{ TNexFilesystem }

function TNexFilesystem.AllocFileBlock(AfterBlock: integer): integer;
var
  CurNext : integer;
begin
  Result := FNexMemory.Sector0F.FirstFree;
  if Result = $EFFF then
    Result := -1
  else
  begin
    // Say that the next block is free
    FNexMemory.Sector0F.FirstFree := FNexMemory.FatTable[Result-1].NextFAT;

    // Remove from "free" chain
    if FNexMemory.Sector0F.FirstFree <> $EFFF then
      FNexMemory.FatTable[FNexMemory.Sector0F.FirstFree-1].PrevFAT := $EFFF;

    // Insert into the "used" chain
    CurNext := FNexMemory.FatTable[AfterBlock-1].NextFAT;
    if CurNext <> $EFFF then
      FNexMemory.FatTable[CurNext-1].PrevFAT := Result;
    FNexMemory.FatTable[Result-1].NextFAT := CurNext;
    FNexMemory.FatTable[Result-1].PrevFAT := AfterBlock;
    if AfterBlock <> -1 then
      FNexMemory.FatTable[AfterBlock-1].NextFAT := Result;
  end;

  NexDoLog('TNexFilesystem.AllocFileBlock: Allocated block ' + IntToStr(Result));
end;

function TNexFilesystem.AllocFileEntry: PFileEntry;
var
  FELoop      : integer;
  MaxSongNum  : integer;
  SongLoop    : integer;
begin
  Result := nil;
  for FELoop := 0 to MAXSONGS - 1 do
  begin
    if FNexMemory.FileTable[FELoop].SongNumber <> $00FF then
    begin
      // Found an unused slot
      Result := @FNexMemory.FileTable[FELoop];

      // Prepare the entry
      MaxSongNum := -1;
      for SongLoop := 0 to MAXSONGS - 1 do
      begin
        if (FNexMemory.FileTable[SongLoop].SongNumber > MaxSongNum) and
          (FNexMemory.FileTable[SongLoop].SongNumber <> $FF) then
        begin
          MaxSongNum := FNexMemory.FileTable[SongLoop].SongNumber;
        end;
      end;

      FillChar(Result^, SizeOf(Result^), 0);
      Result^.SongNumber := MaxSongNum + 1;
      Result^.BlockStart := $EFFF;
      Result^.Unknown2 := $FFFF;
      DateTimeToNEXDate(Result^, Now);
      NexDoLog('TNexFilesystem.AllocFileEntry: Allocated file entry ' + IntToStr(FELoop));
      break;
    end;
  end;
end;

function TNexFilesystem.BytesToCommit: integer;
var
  BlockLoop : integer;
  MemCur    : pointer;
  MemOrig   : pointer;
begin
  Result := 0;

  // Fix up the endians
  EndianFixAll(FNexMemory);

  // Only write the bits that have changed
  MemCur := @FNexMemory;
  MemOrig := @FOrigMem;
  for BlockLoop := 0 to (SizeOf(FNexMemory) div $200) - 1 do
  begin
    if not CompareMem(MemCur, MemOrig, $200) then
      Result := Result + $200;
  end;

  // Undo the fiddling of the endian
  EndianFixAll(FNexMemory);
end;
procedure TNexFilesystem.Commit;
var
  BlockLoop : integer;
  MemCur    : pointer;
  MemOrig   : pointer;
begin
  // Fix up the endians
  EndianFixAll(FNexMemory);

  // Only write the bits that have changed
  MemCur := @FNexMemory;
  MemOrig := @FOrigMem;
  for BlockLoop := 0 to (SizeOf(FNexMemory) div $200) - 1 do
  begin
    if not CompareMem(MemCur, MemOrig, $200) then
    begin
      NexDoLog('TNexFilesystem.Commit: Comitted block ' + IntToStr(BlockLoop));
      WriteFlashRam(0, BlockLoop, $200, MemCur);  // Guessed parameter
    end;
    MemCur := pointer(longword(MemCur) + $200);
    MemOrig := pointer(longword(MemOrig) + $200);
  end;

  // We now have a new "originl" memory
  Move(FNexMemory, FOrigMem, SizeOf(FNexMemory));

  // Undo the fiddling of the endian
  EndianFixAll(FNexMemory);
end;

constructor TNexFilesystem.Create;
var
  TmpStream : TFileStream;
begin
  inherited Create;

  // Load up the raw data
  MultisectorRead(0, 0, SizeOf(FNexMemory), @FNexMemory, $200);
  Move(FNexMemory, FOrigMem, SizeOf(FNexMemory));
{  TmpStream := TFileStream.Create('nexlowmem.dat', fmCreate);
  TmpStream.Write(FNexMemory, SizeOf(FNexMemory));
  TmpStream.Free;}

  // Fix up the endians
  EndianFixAll(FNexMemory);
end;

destructor TNexFilesystem.Destroy;
begin
  inherited Destroy;
end;

function TNexFilesystem.FreeBytes: integer;
var
  CurBlock  : integer;
begin
  Result := 0;
  CurBlock := FNexMemory.Sector0F.FirstFree;
  if CurBlock = $EFFF then CurBlock := -1;
  while CurBlock <> -1 do
  begin
    inc(Result);
    CurBlock := NextBlock(CurBlock);
  end;
  Result := Result * SECTORSIZE * SECTORSPERBLOCK;
end;

procedure TNexFilesystem.FreeFileBlock(BlockNum : integer);
var
  CurPrev : integer;
  CurNext : integer;
begin
  // Remove from the "used" chain
  CurNext := FNexMemory.FatTable[BlockNum-1].NextFAT;
  CurPrev := FNexMemory.FatTable[BlockNum-1].PrevFAT;
  if CurNext <> $EFFF then
    FNexMemory.FatTable[CurNext-1].PrevFAT := CurPrev;
  if CurPrev <> $EFFF then
    FNexMemory.FatTable[CurPrev-1].NextFAT := CurNext;

  // Insert into the "free" chain
  if FNexMemory.Sector0F.FirstFree <> $EFFF then
    FNexMemory.FatTable[FNexMemory.Sector0F.FirstFree-1].PrevFAT := BlockNum;
  FNexMemory.FatTable[BlockNum-1].NextFAT := FNexMemory.Sector0F.FirstFree;
  FNexMemory.FatTable[BlockNum-1].PrevFAT := $EFFF;

  NexDoLog('TNexFilesystem.FreeFileBlock: Freed block ' + IntToStr(BlockNum));
end;

// This function I beleive is overly complex for what actually needs to be done.
// This is because I'm not actually sure how strict the file entry table is, and
// this function should work no matter what requirements the table has.
procedure TNexFilesystem.FreeFileEntry(FE: PFileEntry);
var
  FENum     : integer;
  SongLoop  : integer;
  VSongNum  : integer;
begin
  // Figure out what song number we are freeing
  FENum := (longword(FE) - longword(@FNexMemory.FileTable[0])) div SizeOf(FNexMemory.FileTable[0]);
  VSongNum := FNexMemory.FileTable[FENum].SongNumber;

  // Shuffle the rest of the songs down
  Move(FNexMemory.FileTable[FENum + 1], FNexMemory.FileTable[FENum],
    (MAXSONGS - FENum - 1) * SizeOf(FNexMemory.FileTable[0]));

  // Add an empty song entry to the end
  FillChar(FNexMemory.FileTable[MAXSONGS - 1], SizeOf(FNexMemory.FileTable[MAXSONGS - 1]), $FF);
  FNexMemory.FileTable[MAXSONGS - 1].SongNumber := $FF;

  // Renumber the songs
  for SongLoop := 0 to MAXSONGS - 1 do
  begin
    if (FNexMemory.FileTable[SongLoop].SongNumber > VSongNum) and
      (FNexMemory.FileTable[SongLoop].SongNumber <> $FF) then
    begin
      dec(FNexMemory.FileTable[SongLoop].SongNumber);
    end;
  end;

  NexDoLog('TNexFilesystem.FreeFileEntry: Freed file entry ' + IntToStr(FENum));
end;

function TNexFilesystem.GetFileEntry(FileNum: integer): PFileEntry;
begin
  Result := @FNexMemory.FileTable[FileNum];
end;

function TNexFilesystem.NextBlock(CurBlock: integer): integer;
begin
  Result := FNexMemory.FatTable[CurBlock-1].NextFAT;
  if Result = $EFFF then Result := -1;
end;

function TNexFilesystem.PrevBlock(CurBlock: integer): integer;
begin
  Result := FNexMemory.FatTable[CurBlock-1].PrevFAT;
  if Result = $EFFF then Result := -1;
end;

procedure TNexFilesystem.Revert;
begin
  Move(FOrigMem, FNexMemory, SizeOf(FNexMemory));
end;

function TNexFilesystem.TotalBytes: integer;
var
  CurBlock  : integer;
begin
  Result := 0;
  for CurBlock := 0 to NUMFATITEMS - 1 do
    if FNexMemory.FATTable[CurBlock].NextFAT <> $FFFF then inc(Result);
  Result := SECTORSIZE * SECTORSPERBLOCK * Result;
end;

{ TNexStream }

constructor TNexStream.Create(nfs: TNexFilesystem; SongNum : integer);
begin
  inherited Create;
  FFileSys := nfs;

  // Create a new song if required
  if SongNum = -1 then
    FFileEntry := FFileSys.AllocFileEntry
  else
    FFileEntry := FFileSys.GetFileEntry(SongNum);

  FCurPos := 0;
end;

destructor TNexStream.Destroy;
begin
  inherited Destroy;
end;

function TNexStream.Read(var Buffer; Count: Integer): Longint;
var
  BytesToGo : integer;
  CurBlock  : integer;
  CurSector : integer;
  TmpSector : pointer;
  NumToMove : integer;
  DstPtr    : pointer;
  RealCount : integer;

  procedure NextSector;
  begin
    inc(CurSector);
    if CurSector = SECTORSPERBLOCK then
    begin
      CurSector := 0;
      CurBlock := FFileSys.NextBlock(CurBlock);
    end;
  end;

begin
  // See if there's anything to read
  Result := 0;
  RealCount := Count;
  if RealCount + FCurPos > Size then
    RealCount := Size - FCurPos;
  if RealCount = 0 then exit;

  NexDoLog('TNexStream.Read(' + IntToStr(FCurPos) + ', ' +
    IntToStr(RealCount) + ')');

  // Find the starting block and sector
  DstPtr := @Buffer;
  BytesToGo := FCurPos;

  // Find the starting block
  CurBlock := FFileEntry.BlockStart;
  while BytesToGo > SECTORSPERBLOCK*SECTORSIZE do
  begin
    CurBlock := FFileSys.NextBlock(CurBlock);
    Dec(BytesToGo, SECTORSPERBLOCK*SECTORSIZE);
  end;

  // Find the starting sector
  CurSector := 0;
  while BytesToGo > SECTORSIZE do
  begin
    inc(CurSector);
    Dec(BytesToGo, SECTORSIZE);
  end;

  // If BytesToGo > 0, then we need to do a partial sector read.
  if BytesToGo > 0 then
  begin
    // Player seems to need this ...
    KickPlayer;
    SelectBank((CurBlock*SECTORSPERBLOCK + CurSector) shr 15);

    GetMem(TmpSector, SECTORSIZE);
    ReadFlashRam(
      (CurBlock*SECTORSPERBLOCK + CurSector) shr 15,
      (CurBlock*SECTORSPERBLOCK + CurSector) and $7FFF,
      SECTORSIZE, TmpSector);

    NumToMove := SECTORSIZE - BytesToGo;
    if NumToMove > RealCount then NumToMove := RealCount;
    Move(pointer(longint(TmpSector) + BytesToGo)^, DstPtr^, NumToMove);
    FreeMem(TmpSector);

    BytesToGo := RealCount - NumToMove;
    DstPtr := pointer(longint(DstPtr) + NumToMove);
    if BytesToGo > 0 then NextSector;
  end
  else
    BytesToGo := RealCount;

  // Read as many whole sectors as possible
  while BytesToGo >= SECTORSIZE do
  begin
    // Player seems to need this ...
    KickPlayer;
    SelectBank((CurBlock*SECTORSPERBLOCK + CurSector) shr 15);

    ReadFlashRam(
      (CurBlock*SECTORSPERBLOCK + CurSector) shr 15,
      (CurBlock*SECTORSPERBLOCK + CurSector) and $7FFF,
      SECTORSIZE, DstPtr);
    DstPtr := pointer(longint(DstPtr) + SECTORSIZE);
    dec(BytesToGo, SECTORSIZE);
    if BytesToGo > 0 then NextSector;
  end;

  // If there's still bytes required, do another partial sector read
  if BytesToGo > 0 then
  begin
    // Player seems to need this ...
    KickPlayer;
    SelectBank((CurBlock*SECTORSPERBLOCK + CurSector) shr 15);
  
    GetMem(TmpSector, SECTORSIZE);
    ReadFlashRam(
      (CurBlock*SECTORSPERBLOCK + CurSector) shr 15,
      (CurBlock*SECTORSPERBLOCK + CurSector) and $7FFF,
      SECTORSIZE, TmpSector);
    Move(TmpSector^, DstPtr^, BytesToGo);
    FreeMem(TmpSector);
  end;

  // All done!
  Result := RealCount;
  FCurPos := FCurPos + RealCount;
end;

function TNexStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  case Origin of
    soFromBeginning:
      begin
        FCurPos := Offset;
      end;

    soFromCurrent:
      begin
        FCurPos := FCurPos + Offset;
      end;

    soFromEnd:
      begin
        FCurPos := FFileEntry.Filesize + Offset;
      end;
  end;

  if FCurPos > FFileEntry.Filesize then
    SetSize(FCurPos);
  Result := FCurPos;
end;

procedure TNexStream.SetSize(NewSize: Integer);
var
  CurBlock  : integer;
  BytesToGo : integer;
begin
  CurBlock := FFileEntry.BlockStart;
  BytesToGo := NewSize;
  while BytesToGo > 0 do
  begin
    if CurBlock = $EFFF then
      CurBlock := FFileSys.AllocFileBlock(CurBlock)
    else
      CurBlock := FFileSys.NextBlock(CurBlock);
    Dec(BytesToGo, SECTORSPERBLOCK*SECTORSIZE);
  end;

  while FFileSys.NextBlock(CurBlock) <> $EFFF do
    FFileSys.FreeFileBlock(FFileSys.NextBlock(CurBlock));
end;

// NOT COMPLETE YET!!!
function TNexStream.Write(const Buffer; Count: Integer): Longint;
var
  BytesToGo : integer;
  CurBlock  : integer;
  CurSector : integer;
  TmpSector : pointer;
  NumToMove : integer;
  SrcPtr    : pointer;

  procedure NextSector;
  begin
    inc(CurSector);
    if CurSector = SECTORSPERBLOCK then
    begin
      CurSector := 0;
      CurBlock := FFileSys.NextBlock(CurBlock);
      if CurBlock = $EFFF then
        CurBlock := FFileSys.AllocFileBlock(CurBlock);
    end;
  end;

begin
  Result := 0;
  // See if there's anything to write
  if Count = 0 then exit;

  // Find the starting block and sector
  SrcPtr := @Buffer;
  BytesToGo := FCurPos;

  // Find the starting block
  CurBlock := FFileEntry.BlockStart;
  while BytesToGo > SECTORSPERBLOCK*SECTORSIZE do
  begin
    if CurBlock = $EFFF then
      CurBlock := FFileSys.AllocFileBlock(CurBlock)
    else
      CurBlock := FFileSys.NextBlock(CurBlock);
    Dec(BytesToGo, SECTORSPERBLOCK*SECTORSIZE);
  end;

  // Find the starting sector
  CurSector := 0;
  while BytesToGo > SECTORSIZE do
  begin
    inc(CurSector);
    Dec(BytesToGo, SECTORSIZE);
  end;

  // If BytesToGo > 0, then we need to do a partial sector write (ie: read in
  // the sector, modify, then write out again).
  if BytesToGo > 0 then
  begin
    GetMem(TmpSector, SECTORSIZE);
    ReadFlashRam(
      (CurBlock*SECTORSPERBLOCK + CurSector) shr 15,
      (CurBlock*SECTORSPERBLOCK + CurSector) and $7FFF,
      SECTORSIZE, TmpSector);

    NumToMove := SECTORSIZE - BytesToGo;
    if NumToMove > Count then NumToMove := Count;
    Move(SrcPtr^, pointer(longint(TmpSector) + BytesToGo)^, NumToMove);
    WriteFlashRam(0, CurBlock*SECTORSPERBLOCK + CurSector, SECTORSIZE, TmpSector);
    FreeMem(TmpSector);

    BytesToGo := Count - NumToMove;
    SrcPtr := pointer(longint(SrcPtr) + NumToMove);
    if BytesToGo > 0 then NextSector;
  end
  else
    BytesToGo := Count;

  // Now, do as many full block writes as we can
  while BytesToGo >= SECTORSIZE do
  begin
    WriteFlashRam(0, CurBlock*SECTORSPERBLOCK + CurSector, SECTORSIZE, SrcPtr);
    SrcPtr := pointer(longint(SrcPtr) + SECTORSIZE);
    dec(BytesToGo, SECTORSIZE);
    if BytesToGo > 0 then NextSector;
  end;

  // If there's still bytes required, do another partial sector write
  if BytesToGo > 0 then
  begin
    GetMem(TmpSector, SECTORSIZE);
    ReadFlashRam(
      (CurBlock*SECTORSPERBLOCK + CurSector) shr 15,
      (CurBlock*SECTORSPERBLOCK + CurSector) and $7FFF,
      SECTORSIZE, TmpSector);
    Move(SrcPtr^, TmpSector^, BytesToGo);
    WriteFlashRam(0, CurBlock*SECTORSPERBLOCK + CurSector, SECTORSIZE, TmpSector);
    FreeMem(TmpSector);
  end;

  // All done!
end;

end.
