unit NEXUtils;

interface

uses
  NEXTypes, NEXLowLevel, DateUtils, Sysutils;

procedure EndianFix16P(var In16 : uint16);
function EndianFix16F(In16 : uint16) : uint16;
procedure FillFileTable(var ft: TFileTable);
procedure MultisectorRead(BankNum : byte; BlockNum : word; NumBytes : longint;
  DataPtr : pointer; BlockSize : longint);
function NEXDateToDateTime(const fe: TFileEntry) : TDateTime;
function NEXDateToStr(const fe: TFileEntry) : string;
procedure DateTimeToNEXDate(var fe: TFileEntry; const DateIn : TDateTime);
function GetBitrate(const fe : TFileEntry) : integer;
procedure EndianFixAll(var nlm : TNEXLowMemory);

implementation

procedure EndianFix16P(var In16 : uint16);
begin
  In16 := (In16 shl 8) or (In16 shr 8);
end;

function EndianFix16F(In16 : uint16) : uint16;
begin
  Result := (In16 shl 8) or (In16 shr 8);
end;

procedure MultisectorRead(BankNum : byte; BlockNum : word; NumBytes : longint;
  DataPtr : pointer; BlockSize : longint);
var
  BlockLoop : integer;
  TmpBlock  : pointer;
  TmpPtr    : pointer;
begin
  TmpPtr := DataPtr;
  // Load up the whole blocks first
  for BlockLoop := BlockNum to BlockNum + (NumBytes div BlockSize) - 1 do
  begin
    KickPlayer;
    SelectBank(BankNum);
    ReadFlashRam(BankNum, BlockLoop, BlockSize, TmpPtr);
    TmpPtr := pointer(longword(TmpPtr) + longword(BlockSize));
  end;
  // Load up any partial block
  if (NumBytes mod BlockSize) > 0 then
  begin
    GetMem(TmpBlock, BlockSize);
    KickPlayer;
    SelectBank(BankNum);
    ReadFlashRam(BankNum, BlockNum + (NumBytes div BlockSize), BlockSize, TmpBlock);
    Move(TmpBlock^, TmpPtr^, NumBytes mod BlockSize);
    FreeMem(TmpBlock);
  end;
end;

procedure FillFileTable(var ft: TFileTable);
var
  FixupLoop : integer;
begin
  // Load up the raw data
  MultisectorRead(0, 0, SizeOf(ft), @ft, $200);
  // Fix up the endian stuff
  for FixupLoop := 0 to MAXSONGS - 1 do
  begin
    EndianFix16P(ft[FixupLoop].SongNumber);
    EndianFix16P(ft[FixupLoop].BlockStart);
  end;
end;

function NEXDateToDateTime(const fe: TFileEntry) : TDateTime;
begin
  Result := EncodeDateTime(
    fe.YearLow + fe.YearHigh * 100,
    fe.Month,
    fe.Day,
    fe.Hour,
    fe.Minute,
    0,
    0);
end;

function NEXDateToStr(const fe: TFileEntry) : string;
begin
  Result := DateTimeToStr(NEXDateToDateTime(fe));
end;

procedure DateTimeToNEXDate(var fe: TFileEntry; const DateIn : TDateTime);
var
  AYear   : word;
  AMonth  : word;
  ADay    : word;
  AHour   : word;
  AMin    : word;
  ASec    : word;
  AMs     : word;
begin
  DecodeDateTime(DateIn, AYear, AMonth, ADay, AHour, AMin, ASec, AMs);

  fe.Day := ADay;
  fe.Month := AMonth;
  fe.YearHigh := AYear div 100;
  fe.YearLow := AYear mod 100;
  fe.Hour := AHour;
  fe.Minute := AMin;
end;

function GetBitrate(const fe : TFileEntry) : integer;
begin
  Result := 0;
  case (fe.MP3FrameHdr shr 4) and $0F of
    0: Result := 0;
    1: Result := 32;
    2: Result := 40;
    3: Result := 48;
    4: Result := 56;
    5: Result := 64;
    6: Result := 80;
    7: Result := 96;
    8: Result := 112;
    9: Result := 128;
    10: Result := 160;
    11: Result := 192;
    12: Result := 224;
    13: Result := 256;
    14: Result := 320;
    15: Result := 0;
  end;
  Result := Result * 1024;
end;

procedure EndianFixAll(var nlm : TNEXLowMemory);
var
  FixupLoop : integer;
begin
  // Change the endian on the song numbers and block start numbers
  for FixupLoop := 0 to MAXSONGS - 1 do
  begin
    EndianFix16P(nlm.FileTable[FixupLoop].SongNumber);
    EndianFix16P(nlm.FileTable[FixupLoop].BlockStart);
  end;

  // Change the endian in the fat table
  for FixupLoop := 0 to NUMFATITEMS - 1 do
  begin
    EndianFix16P(nlm.FATTable[FixupLoop].PrevFAT);
    EndianFix16P(nlm.FATTable[FixupLoop].NextFAT);
  end;
end;

end.
